from spb_curate.config import Config

api_base = "https://api.superb-ai.com"
access_key = None
config = Config()
team_name = None
default_http_client = None
bulk_upload_bytes_max = 256000000  # 256MB
bulk_upload_object_max = 100
timeout = 300  # seconds
log_level = None  # Log level options: DEBUG, INFO

from spb_curate import error  # noqa
from spb_curate.curate import *  # noqa
from spb_curate.version import VERSION  # noqa



from __future__ import annotations

import decimal
import json
from typing import Any, Dict, Iterator, List, Optional, Union

from spb_curate import error, util
from spb_curate.abstract.api.resource import (
    CreateResource,
    ModifyResource,
    PaginateResource,
)
from spb_curate.abstract.superb_ai_object import SuperbAIObject
from spb_curate.curate.api import settings
from spb_curate.curate.api.curate import Job
from spb_curate.curate.api.enums import IouType, JobType, Split
from spb_curate.curate.model.annotation_types import (
    AnnotationType,
    BoundingBox,
    Category,
    Cuboid2D,
    Keypoints,
    Polygon,
    Polyline,
    RotatedBox,
)


class Diagnosis(CreateResource, PaginateResource):
    _endpoints = {
        "create": "/curate/model-diagnosis/datasets/{dataset_id}/diagnoses/",
        "fetch": "/curate/model-diagnosis/datasets/{dataset_id}/diagnoses/{id}/",
        "paginate": "/curate/model-diagnosis/datasets/{dataset_id}/diagnoses/",
    }
    _object_type = "diagnosis"

    @classmethod
    def create(
        cls,
        *,
        access_key: Optional[str] = None,
        team_name: Optional[str] = None,
        dataset_id: str,
        model_name: str,
        class_list: List[str],
        iou_type: IouType,
        metadata: Optional[Dict[str, Union[int, float, decimal.Decimal]]] = None,
    ) -> Diagnosis:
        """
        Creates a diagnosis.

        Parameters
        ----------
        dataset_id
            The ID of the dataset to use for the diagnosis.
        model_name
            The name of the model to diagnose.
        class_list
            The list of class names to diagnose.
        iou_type
            The IoU type of the diagnosis.
        metadata
            The metadata associated with the diagnosis.
            Supported fields: ``beta``, ``target_iou``.
        access_key
            An access key for request authentication.
            If provided, overrides the configuration.
        team_name
            A team name for request authentication.
            If provided, overrides the configuration.

        Returns
        -------
            The created diagnosis.

        Raises
        ------
        ConflictError
            When a diagnosis related to the dataset and model already exists.
        """
        endpoint_params = {"dataset_id": dataset_id}
        param_metadata = {
            "beta": 1.0,
            "class_list": class_list,
            "target_iou": 0.5,
        }

        if metadata:
            param_metadata["beta"] = metadata.get("beta", 1.0)
            param_metadata["target_iou"] = metadata.get("target_iou", 0.5)

        params = {
            "metadata": json.dumps(param_metadata),
            "iou_type": iou_type,
            "model_name": model_name,
            "model_source": "external",
        }

        return super(Diagnosis, cls).create(
            access_key=access_key,
            team_name=team_name,
            endpoint_params=endpoint_params,
            headers=None,
            params=params,
        )

    @classmethod
    def fetch(
        cls,
        *,
        access_key: Optional[str] = None,
        team_name: Optional[str] = None,
        dataset_id: str,
        id: str,
    ) -> Diagnosis:
        """
        Fetches a diagnosis.

        Parameters
        ----------
        dataset_id
            The ID of the dataset to fetch the diagnosis from.
        id
            The ID of the diagnosis to fetch.
        access_key
            An access key for request authentication.
            If provided, overrides the configuration.
        team_name
            A team name for request authentication.
            If provided, overrides the configuration.

        Returns
        -------
            The fetched diagnosis.
        """
        endpoint_params = {"dataset_id": dataset_id, "id": id}

        return super(Diagnosis, cls).fetch(
            access_key=access_key,
            team_name=team_name,
            endpoint_params=endpoint_params,
        )

    @classmethod
    def fetch_all(
        cls,
        *,
        access_key: Optional[str] = None,
        team_name: Optional[str] = None,
        dataset_id: str,
    ) -> List[Diagnosis]:
        """
        Fetches diagnoses that match the provided filters.
        If filters are not provided, fetches all diagnoses.

        Parameters
        ----------
        dataset_id
            The ID of the dataset to fetch the diagnosis from.
        access_key
            An access key for request authentication.
            If provided, overrides the configuration.
        team_name
            A team name for request authentication.
            If provided, overrides the configuration.

        Returns
        -------
            Matching diagnoses.
        """
        all_diagnoses = []
        for page in cls.fetch_page_iter(
            access_key=access_key,
            team_name=team_name,
            dataset_id=dataset_id,
        ):
            all_diagnoses.extend(page.get("results", []))
        return all_diagnoses

    @classmethod
    def fetch_all_iter(
        cls,
        *,
        access_key: Optional[str] = None,
        team_name: Optional[str] = None,
        dataset_id: str,
    ) -> Iterator[Diagnosis]:
        """
        Iterates through diagnoses that match the provided filters.
        If filters are not provided, iterates through all diagnoses.

        Parameters
        ----------
        dataset_id
            The ID of the dataset to fetch the diagnosis from.
        access_key
            An access key for request authentication.
            If provided, overrides the configuration.
        team_name
            A team name for request authentication.
            If provided, overrides the configuration.

        Returns
        -------
            The matching diagnosis iterator.

        Yields
        -------
            The next matching diagnosis.
        """
        for fetch_result in Diagnosis.fetch_page_iter(
            access_key=access_key,
            team_name=team_name,
            dataset_id=dataset_id,
        ):
            for diagnosis in fetch_result.get("results", []):
                yield diagnosis

    @classmethod
    def fetch_page(
        cls,
        *,
        access_key: Optional[str] = None,
        team_name: Optional[str] = None,
        dataset_id: str,
        page: int = 1,
        limit: int = 10,
    ) -> Dict[str, Union[int, List[Diagnosis]]]:
        """
        Fetches a page of diagnoses that match the provided filters.
        If filters are not provided, paginates all diagnoses.

        Parameters
        ----------
        dataset_id
            The ID of the dataset to fetch the diagnosis from.
        page
            The page number.
        limit
            The maximum number of diagnoses in a page.
        access_key
            An access key for request authentication.
            If provided, overrides the configuration.
        team_name
            A team name for request authentication.
            If provided, overrides the configuration.

        Returns
        -------
            A page of matching diagnoses.
        """
        endpoint_params = {"dataset_id": dataset_id}
        params = {"size": limit}

        if page:
            params["page"] = page

        return super(Diagnosis, cls).fetch_page(
            access_key=access_key,
            team_name=team_name,
            endpoint_params=endpoint_params,
            headers=None,
            params=params,
        )

    @classmethod
    def fetch_page_iter(
        cls,
        *,
        access_key: Optional[str] = None,
        team_name: Optional[str] = None,
        dataset_id: str,
    ) -> Iterator[Dict[str, Union[int, List[Diagnosis]]]]:
        """
        Iterates through pages of diagnoses that match the provided filters.
        If filters are not provided, paginates all diagnoses.

        Parameters
        ----------
        dataset_id
            The ID of the dataset to fetch the diagnosis from.
        access_key
            An access key for request authentication.
            If provided, overrides the configuration.
        team_name
            A team name for request authentication.
            If provided, overrides the configuration.

        Returns
        -------
            The matching diagnosis page iterator.

        Yields
        -------
            The next page of matching diagnoses.
        """

        page = 1
        page_result = {}
        limit = settings.FETCH_PAGE_LIMIT

        def fetch_result(page: int):
            page_result = cls.fetch_page(
                access_key=access_key,
                team_name=team_name,
                dataset_id=dataset_id,
                page=page,
                limit=limit,
            )
            return page_result

        page_result = fetch_result(page=page)
        yield page_result

        while page * limit < page_result["count"]:
            page += 1
            page_result = fetch_result(page=page)
            yield page_result

    def add_predictions(
        self,
        *,
        access_key: Optional[str] = None,
        team_name: Optional[str] = None,
        predictions: List[Prediction],
        split: Split = Split.VAL,
        asynchronous: bool = True,
    ) -> Job:
        """
        Creates a job that adds newly initialized predictions to the diagnosis.

        Parameters
        ----------
        predictions
            Newly initialized predictions to add.
        split
            The subset data type used for training the model.
        asynchronous
            Whether to immediately return the job after creating it.
            If set to ``False``, the function waits for the job to finish before returning.
        access_key
            An access key for request authentication.
            If provided, overrides the configuration.
        team_name
            A team name for request authentication.
            If provided, overrides the configuration.

        Returns
        -------
            The created job.
        """
        return Prediction.create_bulk(
            access_key=access_key,
            team_name=team_name,
            dataset_id=self.dataset_id,
            diagnosis_id=self.id,
            predictions=predictions,
            split=split,
            asynchronous=asynchronous,
        )

    def refresh(
        self,
        *,
        access_key: Optional[str] = None,
        team_name: Optional[str] = None,
    ) -> None:
        """
        Refreshes the diagnosis.

        Parameters
        ----------
        access_key
            An access key for request authentication.
            If provided, overrides the configuration.
        team_name
            A team name for request authentication.
            If provided, overrides the configuration.
        """
        endpoint_params = {"id": self.id}

        super(Diagnosis, self).refresh(
            access_key=access_key,
            team_name=team_name,
            endpoint_params=endpoint_params,
        )


class Evaluation(SuperbAIObject):
    _object_type = "evaluation"


class _Model(PaginateResource, ModifyResource):
    _endpoints = {
        "modify": "/curate/model-diagnosis/external-models/{id}/",
        "paginate": "/curate/model-diagnosis/external-models/",
    }
    _endpoints_method = {
        "modify": "patch",
    }
    _object_type = "model"

    @classmethod
    def fetch(
        cls,
        *,
        access_key: Optional[str] = None,
        team_name: Optional[str] = None,
    ) -> _Model:
        raise NotImplementedError

    @classmethod
    def fetch_all(
        cls,
        *,
        access_key: Optional[str] = None,
        team_name: Optional[str] = None,
    ) -> List[_Model]:
        """
        Fetches models that match the provided filters.
        If filters are not provided, fetches all models.

        Parameters
        ----------
        access_key
            An access key for request authentication.
            If provided, overrides the configuration.
        team_name
            A team name for request authentication.
            If provided, overrides the configuration.

        Returns
        -------
            Matching models.
        """
        all_models = []
        for page in cls.fetch_page_iter(
            access_key=access_key,
            team_name=team_name,
        ):
            all_models.extend(page.get("results", []))
        return all_models

    @classmethod
    def fetch_all_iter(
        cls,
        *,
        access_key: Optional[str] = None,
        team_name: Optional[str] = None,
    ) -> Iterator[_Model]:
        """
        Iterates through models that match the provided filters.
        If filters are not provided, iterates through all models.

        Parameters
        ----------
        access_key
            An access key for request authentication.
            If provided, overrides the configuration.
        team_name
            A team name for request authentication.
            If provided, overrides the configuration.

        Returns
        -------
            The matching model iterator.

        Yields
        -------
            The next matching models.
        """
        for fetch_result in _Model.fetch_page_iter(
            access_key=access_key,
            team_name=team_name,
        ):
            for model in fetch_result.get("results", []):
                yield model

    @classmethod
    def fetch_page(
        cls,
        *,
        access_key: Optional[str] = None,
        team_name: Optional[str] = None,
        page: int = 1,
        limit: int = 10,
    ) -> Dict[str, Union[int, List[_Model]]]:
        """
        Fetches a page of models that match the provided filters.
        If filters are not provided, paginates all models.

        Parameters
        ----------
        page
            The page number.
        limit
            The maximum number of diagnoses in a page.
        access_key
            An access key for request authentication.
            If provided, overrides the configuration.
        team_name
            A team name for request authentication.
            If provided, overrides the configuration.

        Returns
        -------
            A page of matching diagnoses.
        """
        params = {"size": limit}

        if page:
            params["page"] = page

        return super(_Model, cls).fetch_page(
            access_key=access_key,
            team_name=team_name,
            endpoint_params=None,
            headers=None,
            params=params,
        )

    @classmethod
    def fetch_page_iter(
        cls,
        *,
        access_key: Optional[str] = None,
        team_name: Optional[str] = None,
    ) -> Iterator[Dict[str, Union[int, List[_Model]]]]:
        """
        Iterates through pages of models that match the provided filters.
        If filters are not provided, paginates all models.

        Parameters
        ----------
        access_key
            An access key for request authentication.
            If provided, overrides the configuration.
        team_name
            A team name for request authentication.
            If provided, overrides the configuration.

        Returns
        -------
            The matching model page iterator.

        Yields
        -------
            The next page of matching models.
        """

        page = 1
        page_result = {}
        limit = settings.FETCH_PAGE_LIMIT

        def fetch_result(page: int):
            page_result = cls.fetch_page(
                access_key=access_key,
                team_name=team_name,
                page=page,
                limit=limit,
            )
            return page_result

        page_result = fetch_result(page=page)
        yield page_result

        while page * limit < page_result["count"]:
            page += 1
            page_result = fetch_result(page=page)
            yield page_result

    def modify(
        self,
        *,
        access_key: Optional[str] = None,
        team_name: Optional[str] = None,
        name: Optional[str] = None,
    ) -> None:
        """
        Modifies the model.

        Parameters
        ----------
        name
            The new name for the model.
        access_key
            An access key for request authentication.
            If provided, overrides the configuration.
        team_name
            A team name for request authentication.
            If provided, overrides the configuration.

        Raises
        ------
        ConflictError
            When a model with the provided name already exists.
        """
        if name is None:
            return

        endpoint_params = {"id": self.id}
        params = {"name": name}

        super(_Model, self).modify(
            access_key=access_key,
            team_name=team_name,
            endpoint_params=endpoint_params,
            params=params,
        )


class Prediction(SuperbAIObject):
    _discriminator_map = {
        "prediction_value": (
            "prediction_type",
            {
                "box": BoundingBox,
                "category": Category,
                "cuboid2d": Cuboid2D,
                "keypoint": Keypoints,
                "polygon": Polygon,
                "polyline": Polyline,
                "rbox": RotatedBox,
            },
        )
    }
    _field_initializers = {"prediction_value": "_init_prediction_value"}

    def __init__(
        self,
        *,
        confidence: Union[int, float, decimal.Decimal],
        image_id: str,
        prediction_class: str,
        prediction_value: Union[
            BoundingBox,
            Category,
            Cuboid2D,
            Keypoints,
            Polygon,
            Polyline,
            RotatedBox,
            dict,
            list,
        ],
        prediction_type: Optional[str] = None,
        **params,
    ):
        """
        Initializes a prediction.
        A newly initialized prediction is incomplete and must be added to an prediction.

        Parameters
        ----------
        confidence
            The score of the prediction.
        image_id
            The ID of the image to add the prediction to.
        prediction_class
            The classification of the prediction (e.g. "person", "vehicle").
        prediction_value
            The value of the prediction.
        prediction_type
            The type of the prediction (e.g. "box", "polygon").
            Will be inferred if ``prediction_value`` is an instance of ``AnnotationType``.
        """

        super(Prediction, self).__init__(
            confidence=confidence,
            image_id=image_id,
            prediction_class=prediction_class,
            prediction_value=prediction_value,
            prediction_type=(
                prediction_value._object_type
                if isinstance(prediction_value, AnnotationType)
                else prediction_type
            ),
            **params,
        )

    def _init_volatile_fields(
        self,
        confidence: Union[int, float, decimal.Decimal],
        image_id: str,
        prediction_class: str,
        prediction_value: Union[
            BoundingBox,
            Category,
            Cuboid2D,
            Keypoints,
            Polygon,
            Polyline,
            RotatedBox,
            dict,
            list,
        ],
        prediction_type: Optional[str] = None,
        **params,
    ) -> None:
        super(Prediction, self)._init_volatile_fields(
            confidence=confidence,
            image_id=image_id,
            prediction_class=prediction_class,
            prediction_value=prediction_value,
            prediction_type=prediction_type,
            **params,
        )

        for k, v, is_required in iter(
            [
                ("confidence", confidence, True),
                ("image_id", image_id, True),
                ("prediction_class", prediction_class, True),
                ("prediction_type", prediction_type, True),
                ("prediction_value", prediction_value, True),
            ]
        ):
            if isinstance(k, list):
                util.validate_argument_list(keys=k, values=v, is_required=is_required)
                for paired_i in range(len(k)):
                    self[k[paired_i]] = util.convert_to_superb_ai_object(
                        data=v[paired_i]
                    )
            else:
                util.validate_argument_value(key=k, value=v, is_required=is_required)
                self[k] = util.convert_to_superb_ai_object(data=v)

        if not prediction_type:
            raise error.ValidationError("prediction_type is required")

        if isinstance(prediction_value, (dict, list)) and not isinstance(
            prediction_value, AnnotationType
        ):
            self._init_prediction_value(prediction_type, prediction_value)

    def _init_prediction_value(
        self,
        prediction_type: str,
        prediction_value: Union[
            BoundingBox,
            Category,
            Cuboid2D,
            Keypoints,
            Polygon,
            Polyline,
            RotatedBox,
        ],
        **params,
    ):
        prediction_type_cls = self.get_cls_by_discriminator(
            field="prediction_value", data={"prediction_type": prediction_type}
        )
        self.prediction_value = prediction_type_cls(raw_data=prediction_value)

    @classmethod
    def create_bulk(
        cls,
        *,
        access_key: Optional[str] = None,
        team_name: Optional[str] = None,
        dataset_id: str,
        diagnosis_id: str,
        predictions: List[Prediction],
        split: Split = Split.VAL,
        asynchronous: bool = True,
    ) -> Job:
        """
        Creates a job that adds newly initialized predictions to a diagnosis.

        Parameters
        ----------
        dataset_id
            The ID of the dataset to add the predictions to.
        diagnosis_id
            The ID of the diagnosis to add the predictions to.
        predictions
            Newly initialized predictions to add.
        split
            The subset data type used for training the model.
        asynchronous
            Whether to immediately return the job after creating it.
            If set to ``False``, the function waits for the job to finish before returning.
        access_key
            An access key for request authentication.
            If provided, overrides the configuration.
        team_name
            A team name for request authentication.
            If provided, overrides the configuration.

        Returns
        -------
            The created job.
        """

        raw_predictions = []

        # Ensure the prediction_values are converted to "raw" form
        for prediction in predictions:
            raw_predictions.append(prediction.to_dict_deep())

        uploaded_param = Job._upload_params(
            access_key=access_key,
            team_name=team_name,
            data=raw_predictions,
        )

        job = Job.create(
            access_key=access_key,
            team_name=team_name,
            job_type=JobType.IMPORT_PREDICTIONS,
            param={
                "dataset_id": dataset_id,
                "diagnosis_id": diagnosis_id,
                "predictions": {"param_id": uploaded_param["id"]},
                "split": split,
            },
        )

        if not asynchronous:
            job.wait_until_complete()

        return job


def create_diagnosis(
    *,
    access_key: Optional[str] = None,
    team_name: Optional[str] = None,
    dataset_id: str,
    model_name: str,
    class_list: List[str],
    iou_type: IouType,
    metadata: Optional[Dict[str, Union[int, float, decimal.Decimal]]] = None,
) -> Diagnosis:
    """
    Creates a diagnosis.

    Parameters
    ----------
    dataset_id
        The ID of the dataset to use for the diagnosis.
    model_name
        The name of the model to diagnose.
    class_list
        The list of class names to diagnose.
    iou_type
        The iou type of the diagnosis.
    metadata
        The metadata associated with the diagnosis.
        Supported fields: ``beta``, ``target_iou``.
    access_key
        An access key for request authentication.
        If provided, overrides the configuration.
    team_name
        A team name for request authentication.
        If provided, overrides the configuration.

    Returns
    -------
        The created diagnosis.

    Raises
    ------
    ConflictError
        When a diagnosis for the (dataset, model_name) pair already exists.
    """
    return Diagnosis.create(
        access_key=access_key,
        team_name=team_name,
        dataset_id=dataset_id,
        model_name=model_name,
        class_list=class_list,
        iou_type=iou_type,
        metadata=metadata,
    )


def fetch_available_models(
    *,
    access_key: Optional[str] = None,
    team_name: Optional[str] = None,
) -> Dict[str, Any]:
    return [
        model.to_dict_deep()
        for model in _Model.fetch_all(access_key=access_key, team_name=team_name)
    ]


def modify_model_name(
    *,
    access_key: Optional[str] = None,
    team_name: Optional[str] = None,
    id: str,
    name: str,
) -> Dict[str, Any]:
    """
    Modifies the model.

    Parameters
    ----------
    id
        The id of the model to modify.
    name
        The new name for the model.
    access_key
        An access key for request authentication.
        If provided, overrides the configuration.
    team_name
        A team name for request authentication.
        If provided, overrides the configuration.

    Raises
    ------
    ConflictError
        When a model with the provided name already exists.
    """
    obj = _Model(id=id)
    obj.modify(
        access_key=access_key,
        team_name=team_name,
        name=name,
    )

    return obj.to_dict_deep()



import base64
import calendar
import datetime
import json
import platform
import time
import uuid
from typing import Optional, Union
from urllib.parse import urlencode, urlsplit, urlunsplit

from spb_curate import error, http_client, util
from spb_curate.superb_ai_response import SuperbAIResponse
from spb_curate.version import VERSION


def _encode_datetime(dttime) -> int:
    if dttime.tzinfo and dttime.tzinfo.utcoffset(dttime) is not None:
        utc_timestamp = calendar.timegm(dttime.utctimetuple())
    else:
        utc_timestamp = time.mktime(dttime.timetuple())

    return int(utc_timestamp)


def _api_encode(data: dict):
    for key, value in iter(data.items()):
        if value is None:
            continue
        if isinstance(value, bool):
            yield (key, str(value).lower())
        elif isinstance(value, datetime.datetime):
            yield (key, _encode_datetime(value))
        elif isinstance(value, list):
            for sub_value in value:
                yield (f"{key}[]", sub_value)
        else:
            yield (key, value)


def _build_api_url(*, url: str, query: str):
    scheme, netloc, path, base_query, fragment = urlsplit(url)

    if base_query:
        query = "%s&%s" % (base_query, query)

    return urlunsplit((scheme, netloc, path, query, fragment))


class APIRequestor(object):
    def __init__(
        self,
        *,
        access_key: Optional[str] = None,
        api_base: Optional[str] = None,
        team_name: Optional[str] = None,
    ) -> None:
        import spb_curate

        self.api_base = api_base or spb_curate.api_base
        self.access_key = access_key or spb_curate.access_key
        self.team_name = team_name or spb_curate.team_name

        if spb_curate.default_http_client:
            self._client: http_client.RequestsClient = spb_curate.default_http_client
        else:
            spb_curate.default_http_client = http_client.RequestsClient()
            self._client: http_client.RequestsClient = spb_curate.default_http_client

    def handle_error_response(self, rbody, rcode, resp, rheaders):
        error_type = ""
        try:
            if isinstance(resp, dict):
                error_data = resp.get("detail", "")
                error_type = resp.get("type", "")
            else:
                error_data = str(resp)
        except (KeyError, TypeError):
            raise error.APIError(
                message="Invalid response object from API: %r (HTTP response code "
                "was %d)" % (rbody, rcode),
                http_status=rbody,
                code=rcode,
                error_body=resp,
            )

        raise self.specific_api_error(
            rbody, rcode, resp, rheaders, error_data, error_type
        )

    def specific_api_error(
        self,
        rbody,
        rcode: int,
        resp,
        rheaders,
        error_data: Optional[Union[str, dict]] = None,
        error_type: Optional[str] = None,
    ):
        message = ""

        if error_data:
            message = (
                error_data.get("message")
                if isinstance(error_data, dict)
                else str(error_data)
            )

        util.log_info(
            "Superb AI API error response",
            error_code=rcode,
            error_type=error_type,
            error_message=message,
        )

        if rcode == 400:
            if error_type == "QUERY_SYNTAX":
                return error.QuerySyntaxError(
                    message=message,
                    code=rcode,
                )
            return error.BadRequestError(
                message=message,
                code=rcode,
            )
        elif rcode == 401:
            return error.AuthenticationError(
                message=message,
                code=rcode,
            )
        elif rcode == 404:
            return error.NotFoundError(
                message=message,
                code=rcode,
            )
        elif rcode == 409:
            return error.ConflictError(
                message=message,
                code=rcode,
            )
        elif rcode == 429:
            return error.TooManyRequestsError(
                message=message,
                code=rcode,
            )
        elif rcode == 500:
            return error.SystemError(
                message=message,
                code=rcode,
            )

        return error.APIError(message=message, code=rcode)

    def request(
        self,
        *,
        method: str,
        url: str,
        params: Optional[dict] = None,
        headers: Optional[dict] = None,
    ):
        rbody, rcode, rheaders, my_access_key = self.request_raw(
            method=method.lower(), url=url, params=params, supplied_headers=headers
        )
        resp = self.interpret_response(rbody=rbody, rcode=rcode, rheaders=rheaders)
        return resp, my_access_key

    def request_headers(self, *, access_key: str, team_name: str, method: str) -> dict:
        user_agent = "SPB/v0 PythonBindings/%s" % (VERSION,)

        ua = {"bindings_version": VERSION, "lang": "python", "publisher": "superb-ai"}

        for attr, func in [
            ["lang_version", platform.python_version],
            ["platform", platform.platform],
            ["uname", lambda: " ".join(platform.uname())],
        ]:
            try:
                val = func()
            except Exception:
                val = "(disabled)"
            ua[attr] = val

        headers = {
            "X-SPB-Client-User-Agent": json.dumps(ua),
            "User-Agent": user_agent,
            "X-Api-Key": access_key,
            "X-Tenant-Id": team_name,
            "Authorization": f"Basic {base64.b64encode(f'{team_name}:{access_key}'.encode('utf-8')).decode('utf-8')}",
        }

        if method == "post" or method == "put" or method == "patch":
            headers["Content-Type"] = "application/json"
            headers.setdefault("Idempotency-Key", str(uuid.uuid4()))

        return headers

    def request_raw(
        self,
        *,
        method: str,
        url: str,
        params: Optional[dict] = None,
        supplied_headers: Optional[dict] = None,
    ):
        """
        Mechanism for issuing an API call
        """

        if self.access_key:
            my_access_key = self.access_key
        else:
            from spb_curate import access_key

            my_access_key = access_key

        if my_access_key is None:
            raise error.AuthenticationError("No access key was provided.")

        if self.team_name:
            my_team_name = self.team_name
        else:
            from spb_curate import team_name

            my_team_name = team_name

        if my_team_name is None:
            raise error.AuthenticationError("No team name was provided.")

        abs_url = f"{self.api_base}{url}"

        # prepare headers
        headers = self.request_headers(
            access_key=my_access_key, team_name=my_team_name, method=method
        )
        if supplied_headers is not None:
            for key, value in iter(supplied_headers.items()):
                headers[key] = value

        def encode_params() -> str:
            return urlencode(list(_api_encode(params or {})))

        if method == "get" or method == "delete":
            if params:
                abs_url = _build_api_url(url=abs_url, query=encode_params())
            post_data = None
        elif method == "post" or method == "put" or method == "patch":
            if headers.get("Content-Type", "") == "application/json":
                post_data = json.dumps(params)
            elif headers.get("Content-Type", "") == "application/x-www-form-urlencoded":
                post_data = encode_params()
            else:
                raise error.APIConnectionError(
                    f"Unrecognized header Content-Type {headers.get('Content-Type', '')}."
                )
        else:
            raise error.APIConnectionError(f"Unrecognized HTTP method {method}.")

        util.log_info("Superb AI API Request", method=method, path=abs_url)
        util.log_debug(
            "POST data",
            post_data=post_data,
        )

        rcontent, rcode, rheaders = self._client.request(
            method=method, url=abs_url, headers=headers, post_data=post_data
        )

        util.log_info("Superb AI API response", path=abs_url, response_code=rcode)
        util.log_debug("API response body", body=rcontent)

        return rcontent, rcode, rheaders, my_access_key

    def _should_handle_code_as_error(self, rcode):
        return not 200 <= rcode < 300

    def interpret_response(self, *, rbody, rcode, rheaders) -> SuperbAIResponse:
        try:
            if hasattr(rbody, "decode"):
                rbody = rbody.decode("utf-8")
            resp = SuperbAIResponse(rbody, rcode, rheaders)
        except Exception:
            raise error.APIError(
                "Invalid response body from API: %s "
                "(HTTP response code was %d)" % (rbody, rcode),
                rbody,
                rcode,
                rheaders,
            )
        if self._should_handle_code_as_error(rcode):
            self.handle_error_response(
                rbody=rbody, rcode=rcode, resp=resp.data, rheaders=rheaders
            )
        return resp



from spb_curate.curate.model import annotation_types
from spb_curate.curate.model.annotation_types import (
    BoundingBox,
    Category,
    Contour,
    Cuboid2D,
    Edge,
    Face,
    Keypoint,
    Keypoints,
    KeypointState,
    Path,
    Point,
    Polygon,
    Polyline,
    RotatedBox,
)



from __future__ import annotations

from typing import Dict, Iterator, List, Optional, Union

from spb_curate import api_requestor, error, util
from spb_curate.abstract.superb_ai_object import SuperbAIObject


class APIResource(SuperbAIObject):
    @classmethod
    def _static_request(
        cls,
        *,
        method_: str,
        url_: str,
        access_key: Optional[str] = None,
        api_base: Optional[str] = None,
        team_name: Optional[str] = None,
        params: Optional[dict] = None,
        headers: Optional[dict] = None,
    ) -> SuperbAIObject:
        requestor = api_requestor.APIRequestor(
            access_key=access_key, api_base=api_base, team_name=team_name
        )
        response, access_key = requestor.request(
            method=method_, url=url_, params=params, headers=headers
        )

        return util.convert_to_superb_ai_object(
            data=response,
            access_key=access_key,
            team_name=team_name,
            cls=cls,
        )

    @classmethod
    def fetch(
        cls,
        *,
        access_key: Optional[str] = None,
        team_name: Optional[str] = None,
        endpoint_params: Optional[dict] = None,
        headers: Optional[dict] = None,
        params: Optional[dict] = None,
    ) -> SuperbAIObject:
        """
        Raises
        ------
        ValidationError
            When `endpoint_params` contains `None` or empty strings.
        """
        url = cls.get_endpoint(name="fetch", params=endpoint_params)
        method = cls.get_endpoint_method(name="fetch", default="get")

        return cls._static_request(
            method_=method,
            url_=url,
            access_key=access_key,
            team_name=team_name,
            params=params,
            headers=headers,
        )

    def refresh(
        self,
        *,
        access_key: Optional[str] = None,
        team_name: Optional[str] = None,
        endpoint_params: Optional[dict] = None,
        headers: Optional[dict] = None,
        params: Optional[dict] = None,
    ) -> None:
        url = self.get_endpoint(name="fetch", params=endpoint_params)
        method = self.get_endpoint_method(name="fetch", default="get")

        refreshed_object = self._static_request(
            method_=method,
            url_=url,
            access_key=access_key,
            team_name=team_name,
            params=params,
            headers=headers,
        )

        self.load_from_dict(
            data=refreshed_object,
            access_key=access_key,
            team_name=team_name,
        )


class CreateResource(APIResource):
    @classmethod
    def create(
        cls,
        *,
        access_key: Optional[str] = None,
        team_name: Optional[str] = None,
        endpoint_params: Optional[dict] = None,
        headers: Optional[dict] = None,
        params: Optional[dict] = None,
    ) -> SuperbAIObject:
        """
        Raises
        ------
        ValidationError
            When `endpoint_params` contains `None` or empty strings.
        """
        url = cls.get_endpoint(name="create", params=endpoint_params)
        method = cls.get_endpoint_method(name="create", default="post")

        return cls._static_request(
            method_=method,
            url_=url,
            access_key=access_key,
            team_name=team_name,
            params=params,
            headers=headers,
        )


class DeleteResource(APIResource):
    @classmethod
    def _cls_delete(
        cls,
        access_key: Optional[str] = None,
        team_name: Optional[str] = None,
        endpoint_params: Optional[dict] = None,
        headers: Optional[dict] = None,
        params: Optional[dict] = None,
    ) -> None:
        """
        Raises
        ------
        ValidationError
            When `endpoint_params` contains `None` or empty strings.
        """
        url = cls.get_endpoint(name="delete", params=endpoint_params)
        method = cls.get_endpoint_method(name="delete", default="delete")

        return cls._static_request(
            method_=method,
            url_=url,
            access_key=access_key,
            team_name=team_name,
            headers=headers,
            params=params,
        )

    def delete(
        self,
        *,
        access_key: Optional[str] = None,
        team_name: Optional[str] = None,
        endpoint_params: Optional[dict] = None,
        headers: Optional[dict] = None,
        params: Optional[dict] = None,
    ) -> None:
        """
        Raises
        ------
        ValidationError
            When `endpoint_params` contains `None` or empty strings.
        """
        self._cls_delete(
            access_key=access_key,
            team_name=team_name,
            endpoint_params=endpoint_params,
            headers=headers,
            params=params,
        )
        self.id = None


class PaginateResource(APIResource):
    @classmethod
    def fetch_all(
        cls,
        *,
        access_key: Optional[str] = None,
        team_name: Optional[str] = None,
        endpoint_params: Optional[dict] = None,
        headers: Optional[dict] = None,
        params: Optional[dict] = None,
    ) -> List[SuperbAIObject]:
        """
        Raises
        ------
        ValidationError
            When `endpoint_params` contains `None` or empty strings.
        """
        url = cls.get_endpoint(name="paginate", params=endpoint_params)
        method = cls.get_endpoint_method(name="paginate", default="get")
        get_params = params.copy() if params else {}

        requestor = api_requestor.APIRequestor(
            access_key=access_key, team_name=team_name
        )

        response, access_key = requestor.request(
            method=method, url=url, params=get_params, headers=headers
        )

        response_data = response.data
        response_data_results = response_data.pop("results", [])

        result_objects: List[SuperbAIObject] = util.convert_to_superb_ai_object(
            data=response_data_results,
            access_key=access_key,
            team_name=team_name,
            cls=cls,
        )

        data = util.convert_to_superb_ai_object(
            data=response_data,
            access_key=access_key,
            team_name=team_name,
        )

        data.update({"results": result_objects})

        return data

    @classmethod
    def fetch_all_iter(
        cls,
        *,
        access_key: Optional[str] = None,
        team_name: Optional[str] = None,
        endpoint_params: Optional[dict] = None,
        headers: Optional[dict] = None,
        params: Optional[dict] = None,
    ) -> Iterator[SuperbAIObject]:
        raise NotImplementedError

    @classmethod
    def fetch_page(
        cls,
        *,
        access_key: Optional[str] = None,
        team_name: Optional[str] = None,
        endpoint_params: Optional[dict] = None,
        headers: Optional[dict] = None,
        params: dict,
    ) -> Dict[str, Union[int, List[SuperbAIObject], str]]:
        """
        Raises
        ------
        ValidationError
            When `endpoint_params` contains `None` or empty strings.
        """
        url = cls.get_endpoint(name="paginate", params=endpoint_params)
        method = cls.get_endpoint_method(name="paginate", default="get")
        requestor = api_requestor.APIRequestor(
            access_key=access_key, team_name=team_name
        )

        response, access_key = requestor.request(
            method=method, url=url, params=params, headers=headers
        )

        response_data = response.data
        response_data_results = response_data.pop("results", [])

        result_objects: List[SuperbAIObject] = util.convert_to_superb_ai_object(
            data=response_data_results,
            access_key=access_key,
            team_name=team_name,
            cls=cls,
        )

        data = util.convert_to_superb_ai_object(
            data=response_data,
            access_key=access_key,
            team_name=team_name,
        )

        data.update({"results": result_objects})

        return data

    @classmethod
    def fetch_page_iter(
        cls,
        *,
        access_key: Optional[str] = None,
        team_name: Optional[str] = None,
        endpoint_params: Optional[dict] = None,
        headers: Optional[dict] = None,
        params: dict,
    ) -> Iterator[Dict[str, Union[int, List[SuperbAIObject], str]]]:
        raise NotImplementedError


class ModifyResource(APIResource):
    @classmethod
    def _cls_modify(
        cls,
        access_key: Optional[str] = None,
        team_name: Optional[str] = None,
        endpoint_params: Optional[dict] = None,
        headers: Optional[dict] = None,
        params: Optional[dict] = None,
    ) -> SuperbAIObject:
        """
        Raises
        ------
        ValidationError
            When `endpoint_params` contains `None` or empty strings.
        """
        url = cls.get_endpoint(name="modify", params=endpoint_params)
        method = cls.get_endpoint_method(name="modify", default="put")

        return cls._static_request(
            method_=method,
            url_=url,
            access_key=access_key,
            team_name=team_name,
            headers=headers,
            params=params,
        )

    def modify(
        self,
        *,
        access_key: Optional[str] = None,
        team_name: Optional[str] = None,
        endpoint_params: Optional[dict] = None,
        headers: Optional[dict] = None,
        params: Optional[dict] = None,
    ) -> SuperbAIObject:
        """
        Raises
        ------
        ValidationError
            When `endpoint_params` contains `None` or empty strings.
        """
        modified_object = self._cls_modify(
            access_key=access_key,
            team_name=team_name,
            endpoint_params=endpoint_params,
            headers=headers,
            params=params,
        )

        self.load_from_dict(
            data=modified_object,
            access_key=access_key,
            team_name=team_name,
        )



from spb_curate import util
from spb_curate.abstract import SuperbAIObject


class BaseImageSource(SuperbAIObject):
    def __init__(
        self,
        *,
        type: str,
        **params,
    ):
        super(BaseImageSource, self).__init__(type=type, **params)

        for k, v in iter([("type", type)]):
            self[k] = util.convert_to_superb_ai_object(data=v)



from __future__ import annotations

import decimal
import json
from typing import Dict, List, NewType, Optional, Tuple, Union

from spb_curate import util
from spb_curate.abstract.superb_ai_object import SuperbAIObject


class AnnotationType(SuperbAIObject):
    _property_fields = set(["raw"])
    _raw_data_child_cls = (None, None)
    _raw_data_field_map = {}
    _object_type = ""

    def __init__(
        self,
        **params,
    ):
        """Initialize an annotation type instance. This initialisation
        method will load the data from a raw form if it exists in the
        init parameters.

        Parameters
        ----------
        faces
            used to contruct a polygon
        raw_data
            raw form of the annotation type using nested lists
        """
        raw_data = params.get("raw_data", None)

        super(AnnotationType, self).__init__(**params)
        self.load_from_raw_data(data=raw_data, **params)

    def load_from_raw_data(self, data: Optional[any] = None, **params):
        """Recursively initalise and load a class instance from a 'raw data' form.

        Parameters
        ----------
        data, optional
            the raw data to be loaded into the class instance, by default None
        """
        if data is None:
            return

        prepared_data = params.copy()

        if isinstance(data, dict):
            data = {v: data[k] for k, v in self._raw_data_field_map.items()} or data
            for field_key, item_data in data.items():
                if (
                    field_key in self._field_class_map
                    and self._field_class_map[field_key]
                ):
                    if isinstance(item_data, list):
                        value = [
                            self._field_class_map[field_key](**{"raw_data": x})
                            for x in item_data
                        ]
                    else:
                        value = self._field_class_map[field_key](
                            **{"raw_data": item_data}
                        )
                else:
                    value = item_data

                prepared_data.update({field_key: value})

            self.load_from_dict(data=prepared_data)
        elif isinstance(data, list):
            children_key, field_cls = self._raw_data_child_cls

            if field_cls:
                items: field_cls = []
                for item_data in data:
                    items.append(field_cls(**{"raw_data": item_data}))

                prepared_data.update({children_key: items})
                self.load_from_dict(data=prepared_data)

        if "raw_data" in self:
            del self.raw_data

    @property
    def raw(self):
        raise NotImplemented

    def __str__(self):
        return json.dumps(
            self.raw,
            sort_keys=True,
            indent=2,
            cls=self.ReprJSONEncoder,
        )


NumberType = Union[int, float, decimal.Decimal]


PointRawType = NewType(
    "PointRawType",
    Dict[str, NumberType],
)


class Point(AnnotationType):
    _field_class_map = {"x": None, "y": None}

    def __init__(
        self,
        *,
        x: Optional[NumberType] = None,
        y: Optional[NumberType] = None,
        raw_data: Optional[PointRawType] = None,
        **params,
    ):
        """Initialize a point used in annotations.

        Parameters
        ----------
        x
            x coordinate
        y
            y coordinate
        raw_data
            a raw form of the point type, {"x": 0, "y": 1}
        """
        super(Point, self).__init__(x=x, y=y, raw_data=raw_data, **params)

    def _init_volatile_fields(
        self,
        x: Optional[NumberType] = None,
        y: Optional[NumberType] = None,
        raw_data: Optional[PointRawType] = None,
        **params,
    ) -> None:
        super(Point, self)._init_volatile_fields(
            x=x,
            y=y,
            raw_data=raw_data,
            **params,
        )

        native_args = [("x", x), ("y", y)]
        self._init_volatile_fields_validate(
            key_value_pair_lists=[native_args, ("raw_data", raw_data)]
        )
        self._init_volatile_fields_load(fields=native_args)

    @property
    def raw(self):
        return {"x": self.x, "y": self.y}


class Contour(AnnotationType):
    _field_class_map = {"points": Point}
    _raw_data_child_cls: Tuple[str, AnnotationType] = ("points", Point)

    def __init__(
        self,
        *,
        points: Optional[List[Point]] = None,
        raw_data: Optional[List[PointRawType]] = None,
        **params,
    ):
        """Initialize a Contour used in annotations.

        Parameters
        ----------
        points
            used to construct a contour.
        raw_data
            a raw form of the contour type, [{"x": 0, "y": 1}, {"x": 1, "y": 1}]
        """
        super(Contour, self).__init__(points=points, raw_data=raw_data, **params)

    def _init_volatile_fields(
        self,
        points: Optional[List[Point]] = None,
        raw_data: Optional[List[PointRawType]] = None,
        **params,
    ) -> None:
        super(Contour, self)._init_volatile_fields(
            points=points,
            raw_data=raw_data,
            **params,
        )

        native_args = [("points", points)]
        self._init_volatile_fields_validate(
            key_value_pair_lists=[native_args, ("raw_data", raw_data)]
        )
        self._init_volatile_fields_load(fields=native_args)

    @property
    def raw(self):
        return [point.raw for point in self.points]


class Face(AnnotationType):
    _field_class_map = {"contours": Contour}
    _raw_data_child_cls: Tuple[str, AnnotationType] = ("contours", Contour)

    def __init__(
        self,
        *,
        contours: Optional[List[Contour]] = None,
        raw_data: Optional[List[List[PointRawType]]] = None,
        **params,
    ):
        """Initialize a face used in annotations.

        Parameters
        ----------
        contours
            used to contruct a face
        raw_data
            a raw form of the face type, [[{"x": 0, "y": 1}, {"x": 1, "y": 1}, {"x": 0, "y": 0}]]
        """
        super(Face, self).__init__(contours=contours, raw_data=raw_data, **params)

    def _init_volatile_fields(
        self,
        contours: Optional[List[Contour]] = None,
        raw_data: Optional[List[List[PointRawType]]] = None,
        **params,
    ) -> None:
        super(Face, self)._init_volatile_fields(
            contours=contours,
            raw_data=raw_data,
            **params,
        )

        native_args = [("contours", contours)]
        self._init_volatile_fields_validate(
            key_value_pair_lists=[native_args, ("raw_data", raw_data)]
        )
        self._init_volatile_fields_load(fields=native_args)

    @property
    def raw(self):
        return [contour.raw for contour in self.contours]


class BoundingBox(AnnotationType):
    _object_type = "box"

    def __init__(
        self,
        *,
        x: Optional[NumberType] = None,
        y: Optional[NumberType] = None,
        width: Optional[NumberType] = None,
        height: Optional[NumberType] = None,
        **params,
    ):
        """Initialize a bounding box used for annotations.

        Parameters
        ----------
        x
            x coordinate
        y
            y coordinate
        width
            width of the bounding box
        height
            height of the bounding box
        raw_data
            the raw data form of the BoundingBox, a dictionary of the fields respectively
        """
        super(BoundingBox, self).__init__(
            x=x, y=y, width=width, height=height, **params
        )

    def _init_volatile_fields(
        self,
        x: Optional[NumberType] = None,
        y: Optional[NumberType] = None,
        width: Optional[NumberType] = None,
        height: Optional[NumberType] = None,
        raw_data: Optional[Dict[str, NumberType]] = None,
        **params,
    ) -> None:
        super(BoundingBox, self)._init_volatile_fields(
            x=x,
            y=y,
            width=width,
            height=height,
            raw_data=raw_data,
            **params,
        )

        native_args = [
            ("x", x),
            ("y", y),
            ("width", width),
            ("height", height),
        ]

        self._init_volatile_fields_validate(
            key_value_pair_lists=[native_args, ("raw_data", raw_data)]
        )
        self._init_volatile_fields_load(fields=native_args)

    @property
    def raw(self):
        return self.to_dict()


class KeypointState(AnnotationType):
    _field_class_map = {"visible": None, "valid": None}

    def __init__(
        self,
        *,
        visible: Optional[bool] = None,
        valid: Optional[bool] = None,
        raw_data: Optional[Dict[str, bool]] = None,
        **params,
    ):
        """Possible states for Keypoints.

        Parameters
        ----------
        visible
            Visibility of Keypoint
        valid
            Validity of Keypoint
        raw_data
            a raw form of the KeypointState, {"visible": True, "valid": True}
        """
        super(KeypointState, self).__init__(
            visible=visible, valid=valid, raw_data=raw_data, **params
        )

    def _init_volatile_fields(
        self,
        visible: Optional[bool] = None,
        valid: Optional[bool] = None,
        raw_data: Optional[Dict[str, bool]] = None,
        **params,
    ) -> None:
        super(KeypointState, self)._init_volatile_fields(
            visible=visible,
            valid=valid,
            raw_data=raw_data,
            **params,
        )

        native_args = [("visible", visible), ("valid", valid)]
        self._init_volatile_fields_validate(
            key_value_pair_lists=[native_args, ("raw_data", raw_data)]
        )
        self._init_volatile_fields_load(fields=native_args)

    @property
    def raw(self):
        return {"visible": self.visible, "valid": self.valid}


KeypointStateRawType = NewType("KeypointStateRawType", Dict[str, bool])

KeypointRawType = NewType(
    "KeypointRawType",
    Dict[str, Union[NumberType, KeypointStateRawType, str]],
)


class Keypoint(AnnotationType):
    _field_class_map = {"x": None, "y": None, "visible": None, "name": None}

    def __init__(
        self,
        *,
        name: Optional[str] = None,
        x: Optional[NumberType] = None,
        y: Optional[NumberType] = None,
        state: Optional[KeypointState] = None,
        raw_data: Optional[KeypointRawType] = None,
        **params,
    ):
        """Initialize a Keypoint used in annotations.

        Parameters
        ----------
        name
            name of the Keypoint
        x
            x coordinate
        y
            y coordinate
        state
            state of Keypoint
        raw_data
            a raw form of the point type, {"x": 0, "y": 1, "visible": False, "name": "..."}
        """
        super(Keypoint, self).__init__(
            name=name, x=x, y=y, state=state, raw_data=raw_data, **params
        )

    def _init_volatile_fields(
        self,
        name: Optional[str] = None,
        x: Optional[NumberType] = None,
        y: Optional[NumberType] = None,
        state: Optional[KeypointState] = None,
        raw_data: Optional[KeypointRawType] = None,
        **params,
    ) -> None:
        super(Keypoint, self)._init_volatile_fields(
            name=name,
            x=x,
            y=y,
            state=state,
            raw_data=raw_data,
            **params,
        )

        native_args = [("name", name), ("x", x), ("y", y), ("state", state)]
        self._init_volatile_fields_validate(
            key_value_pair_lists=[native_args, ("raw_data", raw_data)]
        )
        # load the fields from raw_data if it exists there
        native_args = [(k, v) for k, v in raw_data.items()] if raw_data else native_args
        self._init_volatile_fields_load(fields=native_args)

    @property
    def raw(self):
        return self.to_dict()


EdgeRawType = NewType(
    "EdgeRawType",
    Dict[str, int],
)


class Edge(AnnotationType):
    def __init__(
        self,
        *,
        x: Optional[int] = None,
        y: Optional[int] = None,
        raw_data: Optional[EdgeRawType] = None,
        **params,
    ):
        """Initialize an Edge used in annotations.

        Parameters
        ----------
        x
            the first endpoint of the edge
        y
            the second endpoint of the edge
        raw_data
            a raw form of the Edge type, {"x": 0, "y": 1}
        """
        super(Edge, self).__init__(x=x, y=y, raw_data=raw_data, **params)

    def _init_volatile_fields(
        self,
        x: Optional[int] = None,
        y: Optional[int] = None,
        raw_data: Optional[EdgeRawType] = None,
        **params,
    ) -> None:
        super(Edge, self)._init_volatile_fields(
            x=x,
            y=y,
            raw_data=raw_data,
            **params,
        )

        native_args = [("x", x), ("y", y)]
        self._init_volatile_fields_validate(
            key_value_pair_lists=[native_args, ("raw_data", raw_data)]
        )
        self._init_volatile_fields_load(fields=native_args)

    @property
    def raw(self):
        return {"x": self.x, "y": self.y}


KeypointsRawType = NewType(
    "KeypointsRawType", Dict[str, Union[List[KeypointRawType], List[EdgeRawType]]]
)


class Keypoints(AnnotationType):
    _field_class_map = {"points": Keypoint, "edges": Edge}
    _object_type = "keypoint"

    def __init__(
        self,
        *,
        points: Optional[List[Keypoint]] = None,
        edges: Optional[List[Edge]] = None,
        raw_data: Optional[KeypointsRawType] = None,
        **params,
    ):
        """Initialize a Path used in annotations.

        Parameters
        ----------
        points
            used to plot the keypoints.
        edges
            used to construct the keypoint's edges.
        raw_data
            a raw form of the path type	{"points": [{"x": 100, "y": 100, "state": {"visible": True, "valid": True}, "name": "left_eye"}, {"x": 200, "y": 300, "state": {"visible": True, "valid": True}, "name": "right_eye"}, {"state": {"visible": False, "valid": False}, "name": "nose"}], "edges": [{"x": 0, "y": 1}, {"x": 0, "y": 2}]}
        """
        super(Keypoints, self).__init__(
            points=points, edges=edges, raw_data=raw_data, **params
        )

    def _init_volatile_fields(
        self,
        points: Optional[List[Keypoint]] = None,
        edges: Optional[List[Edge]] = None,
        raw_data: Optional[KeypointsRawType] = None,
        **params,
    ) -> None:
        super(Keypoints, self)._init_volatile_fields(
            points=points,
            edges=edges,
            raw_data=raw_data,
            **params,
        )

        native_args = [("points", points), ("edges", edges)]
        self._init_volatile_fields_validate(
            key_value_pair_lists=[native_args, ("raw_data", raw_data)]
        )
        self._init_volatile_fields_load(fields=native_args)

    @property
    def raw(self):
        return {
            "edges": [edge.raw for edge in self.edges],
            "points": [point.raw for point in self.points],
        }


class Path(AnnotationType):
    _field_class_map = {"points": Point}
    _raw_data_child_cls: Tuple[str, AnnotationType] = ("points", Point)

    def __init__(
        self,
        *,
        points: Optional[List[Point]] = None,
        raw_data: Optional[List[PointRawType]] = None,
        **params,
    ):
        """Initialize a Path used in annotations.

        Parameters
        ----------
        points
            used to construct a path.
        raw_data
            a raw form of the path type, [{"x": 0, "y": 1}, {"x": 1, "y": 1}]
        """
        super(Path, self).__init__(points=points, raw_data=raw_data, **params)

    def _init_volatile_fields(
        self,
        points: Optional[List[Point]] = None,
        raw_data: Optional[List[PointRawType]] = None,
        **params,
    ) -> None:
        super(Path, self)._init_volatile_fields(
            points=points,
            raw_data=raw_data,
            **params,
        )

        native_args = [("points", points)]
        self._init_volatile_fields_validate(
            key_value_pair_lists=[native_args, ("raw_data", raw_data)]
        )
        self._init_volatile_fields_load(fields=native_args)

    @property
    def raw(self):
        return [point.raw for point in self.points]


class Polygon(AnnotationType):
    _field_class_map = {"faces": Face}
    _raw_data_field_map = {"points": "faces"}
    _object_type = "polygon"

    def __init__(
        self,
        *,
        faces: Optional[List[Face]] = None,
        raw_data: Optional[Dict[str, List[List[List[PointRawType]]]]] = None,
        **params,
    ):
        """Initialize a polygon used for annotations.

        Parameters
        ----------
        faces
            used to contruct a polygon
        raw_data
            a raw form of the polygon, {"points": [[[{"x": 0, "y": 0}, {"x": 0, "y": 1}, {"x": 1, "y": 1}, {"x": 1, "y": 0}, {"x": 0, "y": 0}]]]}
        """

        super(Polygon, self).__init__(faces=faces, raw_data=raw_data, **params)

    def _init_volatile_fields(
        self,
        faces: Optional[List[Face]] = None,
        raw_data: Optional[Dict[str, List[List[List[PointRawType]]]]] = None,
        **params,
    ) -> None:
        super(Polygon, self)._init_volatile_fields(
            faces=faces,
            raw_data=raw_data,
            **params,
        )

        native_args = [("faces", faces)]
        self._init_volatile_fields_validate(
            key_value_pair_lists=[native_args, ("raw_data", raw_data)]
        )
        self._init_volatile_fields_load(fields=native_args)

    @property
    def raw(self):
        return {"points": [face.raw for face in self.faces]}

    def to_dict(self):
        return self.raw


class Polyline(AnnotationType):
    _field_class_map = {"paths": Path}
    _raw_data_field_map = {"points": "paths"}
    _object_type = "polyline"

    def __init__(
        self,
        *,
        paths: Optional[List[Path]] = None,
        raw_data: Optional[List[List[List[PointRawType]]]] = None,
        **params,
    ):
        """Initialize a polyline used for annotations.

        Parameters
        ----------
        paths
            used to contruct a polyline
        raw_data
            a raw form of the polyline, {"points": [[{"x": 0, "y": 0}, {"x": 0, "y": 1}, {"x": 1, "y": 1}]]}
        """

        super(Polyline, self).__init__(paths=paths, raw_data=raw_data, **params)

    def _init_volatile_fields(
        self,
        paths: Optional[List[Path]] = None,
        raw_data: Optional[List[List[List[PointRawType]]]] = None,
        **params,
    ) -> None:
        super(Polyline, self)._init_volatile_fields(
            paths=paths,
            raw_data=raw_data,
            **params,
        )

        native_args = [("paths", paths)]
        self._init_volatile_fields_validate(
            key_value_pair_lists=[native_args, ("raw_data", raw_data)]
        )
        self._init_volatile_fields_load(fields=native_args)

    @property
    def raw(self):
        return {"points": [path.raw for path in self.paths]}

    def to_dict(self):
        return self.raw


class RotatedBox(AnnotationType):
    _object_type = "rbox"

    def __init__(
        self,
        *,
        cx: Optional[NumberType] = None,
        cy: Optional[NumberType] = None,
        width: Optional[NumberType] = None,
        height: Optional[NumberType] = None,
        angle: Optional[NumberType] = None,
        raw_data: Optional[Dict[str, NumberType]] = None,
        **params,
    ):
        """Initialize a rotated box used for annotations.

        Parameters
        ----------
        cx
            cx coordinate
        cy
            cy coordinate
        width
            width of the rotated box
        height
            height of the rotated box
        angle
            angle of the rotated box
        raw_data
            the raw data form of the RotatedBox a dict of the values
        """
        super(RotatedBox, self).__init__(
            cx=cx,
            cy=cy,
            width=width,
            height=height,
            angle=angle,
            raw_data=raw_data,
            **params,
        )

    def _init_volatile_fields(
        self,
        cx: Optional[NumberType] = None,
        cy: Optional[NumberType] = None,
        width: Optional[NumberType] = None,
        height: Optional[NumberType] = None,
        angle: Optional[NumberType] = None,
        raw_data: Optional[Dict[str, NumberType]] = None,
        **params,
    ) -> None:
        super(RotatedBox, self)._init_volatile_fields(
            cx=cx,
            cy=cy,
            width=width,
            height=height,
            angle=angle,
            raw_data=raw_data,
            **params,
        )

        native_args = [
            ("cx", cx),
            ("cy", cy),
            ("width", width),
            ("height", height),
            ("angle", angle),
        ]

        self._init_volatile_fields_validate(
            key_value_pair_lists=[native_args, ("raw_data", raw_data)]
        )
        self._init_volatile_fields_load(fields=native_args)

    @property
    def raw(self):
        return self.to_dict()


class Cuboid2D(AnnotationType):
    _field_class_map = {"near": BoundingBox, "far": BoundingBox}
    _object_type = "cuboid2d"

    def __init__(
        self,
        *,
        near: Optional[BoundingBox] = None,
        far: Optional[BoundingBox] = None,
        raw_data: Optional[Dict[str, NumberType]] = None,
        **params,
    ):
        """Initialize a 2d cuboid used for annotations.

        Parameters
        ----------
        near
            a bounding box
        far
            a bounding box
        raw_data
            the raw data form of the Cuboid2D, a dict with two BoundingBox like dicts
        """
        super(Cuboid2D, self).__init__(
            near=near,
            far=far,
            raw_data=raw_data,
            **params,
        )

    def _init_volatile_fields(
        self,
        near: Optional[BoundingBox] = None,
        far: Optional[BoundingBox] = None,
        raw_data: Optional[Dict[str, NumberType]] = None,
        **params,
    ) -> None:
        super(Cuboid2D, self)._init_volatile_fields(
            near=near,
            far=far,
            raw_data=raw_data,
            **params,
        )

        native_args = [
            ("near", near),
            ("far", far),
        ]

        self._init_volatile_fields_validate(
            key_value_pair_lists=[native_args, ("raw_data", raw_data)]
        )
        self._init_volatile_fields_load(fields=native_args)

    @property
    def raw(self):
        return self.to_dict()


class Category(AnnotationType):
    _object_type = "category"

    def __init__(
        self,
        *,
        value: Optional[Union[str, List[str]]] = None,
        raw_data: Optional[Dict[str, Union[str, List[str]]]] = None,
        **params,
    ):
        """Initialize a category used for annotations.

        Parameters
        ----------
        value
            the value or list of values
        raw_data
            the raw data form of the Category, a dict with name and value details
        """
        super(Category, self).__init__(
            value=value,
            raw_data=raw_data,
            **params,
        )

    def _init_volatile_fields(
        self,
        value: Optional[Union[str, List[str]]] = None,
        raw_data: Optional[Dict[str, Union[str, List[str]]]] = None,
        **params,
    ) -> None:
        super(Category, self)._init_volatile_fields(
            value=value,
            raw_data=raw_data,
            **params,
        )

        native_args = []

        self._init_volatile_fields_validate(
            key_value_pair_lists=[native_args, ("raw_data", raw_data)]
        )
        # value is optional, so included seperately here
        self._init_volatile_fields_load(fields=native_args + [("value", value)])

    @property
    def raw(self):
        return self.to_dict()



import configparser
import os
from pathlib import Path
from typing import Dict, List, Optional, Tuple

ENV_ACCESS_KEY = "SPB_ACCESS_KEY"
ENV_PROFILE = "SPB_PROFILE"
ENV_TEAM_NAME = "SPB_TEAM_NAME"


class Config:
    access_key = None
    profile = None
    team_name = None

    def __init__(self, profile: Optional[str] = None) -> None:
        self.profile = profile or os.getenv(ENV_PROFILE, "default")
        self._set_credentials(profile=profile or os.getenv(ENV_PROFILE, "default"))

    def _read_config(self, *, credential_path: Path, profile: str) -> Dict[str, str]:
        config = configparser.ConfigParser()
        config.read(filenames=credential_path)
        config_key_map = {"account_name": "team_name"}

        def parse_creds(*, keys: List[str]) -> Optional[Dict[str, str]]:
            credential = {}

            for k in keys:
                try:
                    credential[config_key_map.get(k, k)] = config.get(profile, k)
                except (configparser.NoSectionError, configparser.NoOptionError):
                    return None

            return credential

        credential = parse_creds(keys=["team_name", "access_key"])

        if credential is None:
            credential = parse_creds(keys=["access_key", "account_name"])

        return credential

    def _set_credentials(self, *, profile: str) -> None:
        self.access_key = os.getenv(ENV_ACCESS_KEY)
        self.team_name = os.getenv(ENV_TEAM_NAME)

        if not self.access_key or not self.team_name:
            config_path = Path.home().joinpath(".spb", "config")

            if config_path.exists():
                credential = self._read_config(
                    credential_path=config_path, profile=self.profile
                )

                self.access_key = self.access_key or credential.get("access_key", None)
                self.team_name = self.team_name or credential.get("team_name", None)



from spb_curate import curate

OBJECT_MAPPING = {
    curate.Annotation._object_type: curate.Annotation,
    curate.Dataset._object_type: curate.Dataset,
    curate.Image._object_type: curate.Image,
    curate.Job._object_type: curate.Job,
    curate.Slice._object_type: curate.Slice,
}



import textwrap
import threading

import requests
from requests.adapters import HTTPAdapter, Retry

from spb_curate import error

BACKOFF_FACTOR = 1
MAX_RETRY_COUNT = 2
STATUS_FORCE_LIST = (500, 502, 503, 504)
STATUS_FORCE_SET = {x for x in STATUS_FORCE_LIST}


def calculate_backoff(attempt: int, factor: int = BACKOFF_FACTOR) -> int:
    return factor * (2 ** (attempt - 1))


class RequestsClient(object):
    name = "requests"

    def __init__(self, timeout: int = 80, session=None, **kwargs):
        super(RequestsClient, self).__init__(**kwargs)
        self._thread_local = threading.local()
        self._session = session
        self._timeout = timeout

    def request(self, *, method: str, url: str, headers: dict, post_data=None):
        return self._request_internal(
            method=method,
            url=url,
            headers=headers,
            post_data=post_data,
        )

    def _request_internal(self, *, method: str, url: str, headers: dict, post_data):
        kwargs = {}

        if getattr(self._thread_local, "session", None) is None:
            if self._session is None:
                self._session = requests.Session()
                allowed_methods = list(Retry.DEFAULT_ALLOWED_METHODS) + [
                    "POST",
                    "PATCH",
                ]
                retries = Retry(
                    total=MAX_RETRY_COUNT,
                    read=MAX_RETRY_COUNT,
                    connect=MAX_RETRY_COUNT,
                    backoff_factor=BACKOFF_FACTOR,
                    status_forcelist=STATUS_FORCE_LIST,
                    allowed_methods=allowed_methods,
                )
                adapter = HTTPAdapter(max_retries=retries)
                self._session.mount("http://", adapter)
                self._session.mount("https://", adapter)

            self._thread_local.session = self._session

        try:
            try:
                result = self._thread_local.session.request(
                    method,
                    url,
                    headers=headers,
                    data=post_data,
                    timeout=self._timeout,
                    **kwargs,
                )
            except TypeError as e:
                raise TypeError(
                    "Warning: It looks like your installed version of the "
                    '"requests" library is not compatible. Most likely '
                    'the "requests" library is out of date. You can fix '
                    'that by running "pip install -U requests".) The '
                    "request error was: %s" % (e,)
                )

            content = result.content

            status_code = result.status_code
        except Exception as e:
            # Would catch just requests.exceptions.RequestException, but can
            # also raise ValueError, RuntimeError, etc.
            self._handle_request_error(e)
        return content, status_code, result.headers

    def _handle_request_error(self, e):
        # Catch SSL error first as it belongs to ConnectionError,
        # but we don't want to retry
        if isinstance(e, requests.exceptions.SSLError):
            msg = (
                "Could not verify Superb AI's SSL certificate. Please make "
                "sure that your network is not intercepting certificates."
            )
            err = "%s: %s" % (type(e).__name__, str(e))
            should_retry = False
        # Retry only timeout and connect errors; similar to urllib3 Retry
        elif isinstance(
            e,
            (requests.exceptions.Timeout, requests.exceptions.ConnectionError),
        ):
            msg = "Unexpected error communicating with Superb AI."
            err = "%s: %s" % (type(e).__name__, str(e))
            should_retry = True
        # Reached maximum retry attempts
        elif isinstance(e, requests.exceptions.RetryError):
            msg = "Unexpected error communicating with Superb AI."
            err = "%s: %s" % (type(e).__name__, str(e))
            should_retry = False
        # Catch remaining request exceptions
        elif isinstance(e, requests.exceptions.RequestException):
            msg = "Unexpected error communicating with Superb AI."
            err = "%s: %s" % (type(e).__name__, str(e))
            should_retry = False
        else:
            msg = (
                "Unexpected error communicating with Superb AI. "
                "It looks like there's probably a configuration "
                "issue locally."
            )
            err = "A %s was raised" % (type(e).__name__,)
            if str(e):
                err += " with error message %s" % (str(e),)
            else:
                err += " with no error message"
            should_retry = False

        msg = textwrap.fill(msg) + "\n\n(Network error: %s)" % (err,)
        raise error.APIConnectionError(msg, should_retry=should_retry)

    def close(self):
        if getattr(self._thread_local, "session", None) is not None:
            self._thread_local.session.close()



from enum import Enum


class IouType(str, Enum):
    """
    Types of IoU.
    """

    BOX = "bbox"
    POLYGON = "segm"


class JobType(Enum):
    """
    Available types of a job.
    """

    ANNOTATION_IMPORT = "ANNOTATION_IMPORT"
    DELETE_IMAGES = "DELETE_IMAGES"
    IMAGE_IMPORT = "IMAGE_IMPORT"
    IMPORT_PREDICTIONS = "IMPORT_PREDICTIONS"
    UPDATE_SLICE = "UPDATE_SLICE"
    UPDATE_SLICE_BY_QUERY = "UPDATE_SLICE_BY_QUERY"

    def __str__(self):
        return self.value


class Split(str, Enum):
    """
    Types of Split.
    """

    TRAIN = "TRAIN"
    VAL = "VAL"


class SearchFieldMappingType(str, Enum):
    ANNOTATION_CLASS = "annotations.class_count"
    ANNOTATION_METADATA = "annotations.metadata"
    IMAGE_METADATA = "images.metadata"

    def __str__(self):
        return self.value


class SupportedImageFormat(str, Enum):
    BMP = "BMP"
    JPG = "JPG"
    JPEG = "JPEG"
    PNG = "PNG"
    MPO = "MPO"
    WEBP = "WEBP"



from __future__ import annotations

import json
import time
from typing import Dict, Iterator, List, Optional, Union

import requests

from spb_curate import api_requestor, error, util
from spb_curate.abstract.api.resource import PaginateResource
from spb_curate.curate.api import settings
from spb_curate.curate.api.enums import JobType


class Job(PaginateResource):
    _endpoints = {
        "bulk_create": "/curate/batch/jobs/",
        "bulk_create_upload": "/curate/batch/params/",
        "fetch": "/curate/batch/jobs/{id}/",
        "paginate": "/curate/batch/jobs/",
    }
    _object_type = "job"

    @classmethod
    def create(
        cls,
        *,
        access_key: Optional[str] = None,
        team_name: Optional[str] = None,
        job_type: JobType,
        param: dict,
    ) -> Job:
        """
        Creates a job.

        Parameters
        ----------
        job_type
            The type of the job to create.
        param
            The parameters for the job.
            Differs by each ``JobType``.
        access_key
            An access key for request authentication.
            If provided, overrides the configuration.
        team_name
            A team name for request authentication.
            If provided, overrides the configuration.

        Returns
        -------
            The created job.
        """

        if not isinstance(job_type, JobType):
            raise error.ValidationError(f"Invalid job type {job_type}.")

        # Submit the bulk job
        bulk_params = {"job_type": job_type.value, "param": param}
        url = cls.get_endpoint(name="bulk_create", params=None)

        return Job._static_request(
            method_="post",
            url_=url,
            access_key=access_key,
            team_name=team_name,
            params=bulk_params,
            headers=None,
        )

    @classmethod
    def fetch(
        cls,
        *,
        access_key: Optional[str] = None,
        team_name: Optional[str] = None,
        id: str,
    ) -> Job:
        """
        Fetches a job.

        Parameters
        ----------
        id
            The ID of the job to fetch.
        access_key
            An access key for request authentication.
            If provided, overrides the configuration.
        team_name
            A team name for request authentication.
            If provided, overrides the configuration.

        Returns
        -------
            The fetched job.
        """
        endpoint_params = {"id": id}

        return super(Job, cls).fetch(
            access_key=access_key,
            team_name=team_name,
            endpoint_params=endpoint_params,
            headers=None,
            params=None,
        )

    @classmethod
    def fetch_all(
        cls,
        *,
        access_key: Optional[str] = None,
        team_name: Optional[str] = None,
        from_date: Optional[str] = None,
    ) -> List[Job]:
        """
        Fetches jobs that match the date filter.
        If not provided, fetches all jobs.

        Parameters
        ----------
        from_date
            ISO 8601 formatted UTC date string to filter jobs created on or after the specified date.
        access_key
            An access key for request authentication.
            If provided, overrides the configuration.
        team_name
            A team name for request authentication.
            If provided, overrides the configuration.

        Returns
        -------
            Matching jobs.
        """
        all_jobs = []
        for page in cls.fetch_page_iter(
            access_key=access_key,
            team_name=team_name,
            from_date=from_date,
        ):
            all_jobs.extend(page.get("results", []))
        return all_jobs

    @classmethod
    def fetch_all_iter(
        cls,
        *,
        access_key: Optional[str] = None,
        team_name: Optional[str] = None,
        from_date: Optional[str] = None,
    ) -> Iterator[Job]:
        """
        Iterates through jobs that match the date filter.
        If not provided, iterates through all jobs.

        Parameters
        ----------
        from_date
            ISO 8601 formatted UTC date string to filter jobs created on or after the specified date.
        access_key
            An access key for request authentication.
            If provided, overrides the configuration.
        team_name
            A team name for request authentication.
            If provided, overrides the configuration.

        Returns
        -------
            The matching job iterator.

        Yields
        -------
            The next matching job.
        """
        for fetch_result in cls.fetch_page_iter(
            access_key=access_key,
            team_name=team_name,
            from_date=from_date,
        ):
            for job in fetch_result.get("results", []):
                yield job

    @classmethod
    def fetch_page(
        cls,
        *,
        access_key: Optional[str] = None,
        team_name: Optional[str] = None,
        from_date: Optional[str] = None,
        cursor: Optional[str] = None,
        limit: int = 10,
    ) -> Dict[str, Union[int, List[Job]]]:
        """
        Fetches a page of jobs that match the date filter.
        If not provided, paginates all jobs.

        Parameters
        ----------
        cursor
            A cursor for pagination.
            Pass in `next_cursor` from the previous page to fetch the next page.
        limit
            The maximum number of jobs in a page.
        from_date
            ISO 8601 formatted UTC date string to filter jobs created on or after the specified date.
        access_key
            An access key for request authentication.
            If provided, overrides the configuration.
        team_name
            A team name for request authentication.
            If provided, overrides the configuration.

        Returns
        -------
            A page of matching jobs.
        """
        params = {"limit": limit}

        if from_date:
            params.update({"from_date": from_date})

        if cursor:
            params["cursor"] = cursor

        return super(Job, cls).fetch_page(
            access_key=access_key,
            team_name=team_name,
            endpoint_params=None,
            headers=None,
            params=params,
        )

    @classmethod
    def fetch_page_iter(
        cls,
        *,
        access_key: Optional[str] = None,
        team_name: Optional[str] = None,
        from_date: Optional[str] = None,
    ) -> Iterator[Dict[str, Union[int, List[Job]]]]:
        """
        Iterates through pages of jobs that match the date filter.
        If not provided, paginates all jobs.

        Parameters
        ----------
        from_date
            ISO 8601 formatted UTC date string to filter jobs created on or after the specified date.
        access_key
            An access key for request authentication.
            If provided, overrides the configuration.
        team_name
            A team name for request authentication.
            If provided, overrides the configuration.

        Returns
        -------
            The matching job page iterator.

        Yields
        -------
            The next page of matching jobs.
        """

        cursor = None
        page_result = {}
        limit = settings.FETCH_PAGE_LIMIT

        def fetch_result(cursor: Optional[str] = None):
            page_result = cls.fetch_page(
                access_key=access_key,
                team_name=team_name,
                from_date=from_date,
                cursor=cursor,
                limit=limit,
            )
            return page_result

        page_result = fetch_result(cursor=cursor)
        yield page_result

        while (
            len(page_result.get("results", [])) == limit
            and page_result.get("next_cursor", None) is not None
        ):
            page_result = fetch_result(cursor=page_result.get("next_cursor"))
            yield page_result

    @classmethod
    def _upload_params(
        cls,
        *,
        access_key: Optional[str] = None,
        team_name: Optional[str] = None,
        data: Union[dict, list],
    ) -> dict:
        # Upload params to S3 bucket
        params_data = json.dumps(data, cls=cls.ReprJSONEncoder).encode("UTF-8")
        url = cls.get_endpoint(name="bulk_create_upload", params=None)

        requestor = api_requestor.APIRequestor(
            access_key=access_key, team_name=team_name
        )
        response, access_key = requestor.request(
            method="post",
            url=url,
            params={"file_size": len(params_data)},
            headers=None,
        )

        upload_params_dict = util.convert_to_superb_ai_object(
            data=response, access_key=access_key, team_name=team_name
        )

        put_params_response = requests.put(
            upload_params_dict["upload_url"], data=params_data
        )

        if put_params_response.status_code != 200:
            raise error.SuperbAIError(
                "There was an error in uploading the job parameters."
            )

        return upload_params_dict

    def refresh(
        self,
        *,
        access_key: Optional[str] = None,
        team_name: Optional[str] = None,
    ) -> None:
        """
        Refreshes the job.

        Parameters
        ----------
        access_key
            An access key for request authentication.
            If provided, overrides the configuration.
        team_name
            A team name for request authentication.
            If provided, overrides the configuration.
        """
        endpoint_params = {"id": self.id}

        super(Job, self).refresh(
            access_key=access_key, team_name=team_name, endpoint_params=endpoint_params
        )

    def wait_until_complete(self, *, timeout: Optional[int] = None) -> None:
        """
        Waits until the job has either completed or failed.

        Parameters
        ----------
        timeout
            The maximum time in seconds to wait for.
            If not provided, ``spb_curate.timeout`` will be used.
        """
        from spb_curate import timeout as default_timeout

        if timeout is None:
            timeout = default_timeout

        frequency = 2  # seconds

        while True:
            if self.status in ["FAILED", "COMPLETE"]:
                break

            wait_seconds = min(timeout, frequency)

            if wait_seconds <= 0:
                break

            timeout -= frequency
            time.sleep(wait_seconds)

            self.refresh()
